from wtforms.fields.core import Field
from wtforms.utils import UnsetValue
from wtforms.validators import InputRequired
from wtforms.widgets.core import TextInput

from flask_wtf import FlaskForm
from wtforms import StringField, FloatField, HiddenField, BooleanField, IntegerField
import inspect


def find_variable(data, script):
    # TODO: needs to check for valid order of variables, important when editting
    added_variables: list[dict[str, str]] = [action for action in script.currently_editing_script if
                                             action["instrument"] == "variable"
                                             # or action["return"] # TODO find returns
                                             ]
    for added_variable in added_variables:
        if added_variable["action"] == data:
            return data, added_variable["args"]
        # if added_variable["return"] == data:
        #     return data, None
    return None, None


class VariableOrStringField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrStringField, self).__init__(label, validators, **kwargs)
        self.script = script

    def process_formdata(self, valuelist):
        if valuelist:
            if not self.script.editing_type == "script" and valuelist[0].startswith("#"):
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]

    def _value(self):
        if self.script:
            variable, value = find_variable(self.data, self.script)
            if variable:
                return variable

        return str(self.data) if self.data is not None else ""


class VariableOrFloatField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrFloatField, self).__init__(label, validators, **kwargs)
        self.script = script

    def _value(self):
        if self.script:
            variable, value = find_variable(self.data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return self.raw_data[0]
        if self.data is not None:
            return str(self.data)
        return ""

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        elif valuelist[0].startswith("#"):
            if not self.script.editing_type == "script":
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]
            return
        try:
            if self.script:
                try:
                    variable, value = find_variable(valuelist[0], self.script)
                    if variable:
                        float(value)
                        self.data = str(variable)
                        return
                except ValueError:
                    pass

            self.data = float(valuelist[0])
        except ValueError as exc:
            self.data = None
            raise ValueError(self.gettext("Not a valid float value.")) from exc


unset_value = UnsetValue()


class VariableOrIntField(Field):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrIntField, self).__init__(label, validators, **kwargs)
        self.script = script

    def _value(self):
        if self.script:
            variable, value = find_variable(self.data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return self.raw_data[0]
        if self.data is not None:
            return str(self.data)
        return ""

    # def process_data(self, value):
    #
    #     if self.script:
    #         variable, var_value = find_variable(value, self.script)
    #         if variable:
    #             try:
    #                 int(var_value)
    #                 self.data = str(variable)
    #                 return
    #             except ValueError:
    #                 pass
    #     if value is None or value is unset_value:
    #         self.data = None
    #         return
    #     try:
    #         self.data = int(value)
    #     except (ValueError, TypeError) as exc:
    #         self.data = None
    #         raise ValueError(self.gettext("Not a valid integer value.")) from exc

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        if self.script:
            variable, var_value = find_variable(valuelist[0], self.script)
            if variable:
                try:
                    int(var_value)
                    self.data = str(variable)
                    return
                except ValueError:
                    pass
        if valuelist[0].startswith("#"):
            if not self.script.editing_type == "script":
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]
            return
        try:
            self.data = int(valuelist[0])
        except ValueError as exc:
            self.data = None
            raise ValueError(self.gettext("Not a valid integer value.")) from exc


class VariableOrBoolField(BooleanField):
    widget = TextInput()

    def __init__(self, label='', validators=None, script=None, **kwargs):
        super(VariableOrBoolField, self).__init__(label, validators, **kwargs)
        self.script = script

    def process_data(self, value):

        if self.script:
            variable, var_value = find_variable(value, self.script)
            if variable:
                try:
                    bool(var_value)
                    return variable
                except ValueError:
                    return

        self.data = bool(value)

    def process_formdata(self, valuelist):
        if not valuelist or type(valuelist) is list and valuelist[0] == '':
            self.data = False
        elif valuelist and valuelist[0].startswith("#"):
            if not self.script.editing_type == "script":
                raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
            self.data = valuelist[0]
        else:
            self.data = True

    def _value(self):

        if self.script:
            variable, value = find_variable(self.raw_data, self.script)
            if variable:
                return variable

        if self.raw_data:
            return str(self.raw_data[0])
        return "y"


def format_name(name):
    """Converts 'example_name' to 'Example Name'."""
    name = name.split(".")[-1]
    text = ' '.join(word for word in name.split('_'))
    return text.capitalize()


def create_form_for_method(method, method_name, autofill, script=None, design=True):
    class DynamicForm(FlaskForm):
        pass

    annotation_mapping = {
        int: (VariableOrIntField if design else IntegerField, 'Enter integer value'),
        float: (VariableOrFloatField if design else FloatField, 'Enter numeric value'),
        str: (VariableOrStringField if design else StringField, 'Enter text'),
        bool: (VariableOrBoolField if design else BooleanField, 'Enter bool value')
    }
    sig = method if type(method) is inspect.Signature else inspect.signature(method)

    for param in sig.parameters.values():
        if param.name == 'self':
            continue
        formatted_param_name = format_name(param.name)
        field_kwargs = {
            "label": formatted_param_name,
            "default": f'#{param.name}' if autofill else (param.default if param.default is not param.empty else ""),
            **({"script": script} if (autofill or design) else {})
        }
        field_class, placeholder_text = annotation_mapping.get(
            param.annotation,
            (VariableOrStringField, f'Enter {param.annotation} value')
        )
        render_kwargs = {"placeholder": placeholder_text}

        # Create the field with additional rendering kwargs for placeholder text
        field = field_class(**field_kwargs, render_kw=render_kwargs)
        setattr(DynamicForm, param.name, field)

    # setattr(DynamicForm, f'add', fname)
    return DynamicForm


# Create forms for each method in DummySDLDeck
def create_add_form(attr, attr_name, autofill, script=None, design=True):
    dynamic_form = create_form_for_method(attr, attr_name, autofill, script, design)
    if design:
        return_value = StringField(label='Save value as', render_kw={"placeholder": "Optional"})
        setattr(dynamic_form, 'return', return_value)
    hidden_method_name = HiddenField(name=f'hidden_name', render_kw={"value": f'{attr_name}'})
    setattr(dynamic_form, 'hidden_name', hidden_method_name)
    return dynamic_form


def create_form_from_module(sdl_module, autofill: bool, script=None, design=True):
    # sdl_deck = DummySDLDeck(DummyPump("COM1"), DummyBalance("COM2"))
    method_forms = {}
    for attr_name in dir(sdl_module):
        attr = getattr(sdl_module, attr_name)
        if inspect.ismethod(attr) and not attr_name.startswith('_'):
            form_class = create_add_form(attr, attr_name, autofill, script, design)
            method_forms[attr_name] = form_class()
    return method_forms


def create_form_from_pseudo(pseudo: dict, autofill: bool, script=None, design=True):
    '''{'dose_liquid': < Signature(amount_in_ml: float, rate_ml_per_minute: float) >}'''
    method_forms = {}
    for attr_name, signature in pseudo.items():
        form_class = create_add_form(signature, attr_name, autofill, script, design)
        method_forms[attr_name] = form_class()
    return method_forms


def create_builtin_form(logic_type):
    class BuiltinFunctionForm(FlaskForm):
        pass

    placeholder_text = f'Enter numbers' if logic_type == 'wait' else f'Enter statement'
    description_text = f'Your variable can be numbers, boolean (True or False) or text ("text")' if logic_type == 'variable' else ''
    field_class = FloatField if logic_type == 'wait' else StringField  # Default to StringField as a fallback
    field_kwargs = {
        "label": f'statement',
        "validators": [InputRequired()] if logic_type in ['wait', "variable"] else [],
        "description": description_text,
    }
    render_kwargs = {"placeholder": placeholder_text}
    field = field_class(**field_kwargs, render_kw=render_kwargs)
    setattr(BuiltinFunctionForm, "statement", field)
    if logic_type == 'variable':
        variable_field = StringField(label=f'variable', validators=[InputRequired()],
                                     description="Your variable name cannot include space",
                                     render_kw=render_kwargs)
        setattr(BuiltinFunctionForm, "variable", variable_field)
    hidden_field = HiddenField(name=f'builtin_name', render_kw={"value": f'{logic_type}'})
    setattr(BuiltinFunctionForm, "builtin_name", hidden_field)
    return BuiltinFunctionForm()


def create_action_button(s: dict):
    style = ""
    if s['instrument'] in ['if', 'while']:
        text = f"{s['action']} {s['args']}"
        style = "background-color: tomato"
    elif s['instrument'] == 'variable':
        text = f"{s['action']} = {s['args']}"
    else:
        # regular action button
        prefix = f"{s['return']} = " if s['return'] else ""
        action_text = f"{s['instrument'].split('.')[-1] if s['instrument'].startswith('deck') else s['instrument']}.{s['action']}"
        arg_string = ""
        if s['args']:
            if type(s['args']) is dict:
                arg_string = "(" + ", ".join([f"{k} = {v}" for k, v in s['args'].items()]) + ")"
            else:
                arg_string = f"= {s['args']}"

        text = f"{prefix}{action_text}  {arg_string}"
    return dict(label=text, style=style, uuid=s["uuid"], id=s["id"])



import inspect
import importlib.util
import os
import pickle
import datetime
import logging
import importlib
import subprocess
import sys

from typing import Optional, Dict, Tuple

from flask import session
from flask_socketio import SocketIO

from .db_models import Script

stypes = ['prep', 'script', 'cleanup']


def get_script_file():
    session_script = session.get("scripts")
    if session_script:
        s = Script()
        s.__dict__.update(**session_script)
        return s
    else:
        return Script(author=session.get('user'))


def post_script_file(script, is_dict=False):
    if is_dict:
        session['scripts'] = script
    else:
        session['scripts'] = script.as_dict()


def create_gui_dir(parent_path):
    os.makedirs(parent_path, exist_ok=True)
    for path in ["config_csv", "scripts", "results", "pseudo_deck"]:
        os.makedirs(os.path.join(parent_path, path), exist_ok=True)


def save_to_history(filepath, history_path):
    connections = []
    try:
        with open(history_path, 'r') as file:
            lines = file.read()
            connections = lines.split('\n')
    except FileNotFoundError:
        pass
    if filepath not in connections:
        with open(history_path, 'a') as file:
            file.writelines(f"{filepath}\n")


def import_history(history_path):
    connections = []
    try:
        with open(history_path, 'r') as file:
            lines = file.read()
            connections = lines.split('\n')
    except FileNotFoundError:
        pass
    connections = [i for i in connections if not i == '']
    return connections


def available_pseudo_deck(path):
    import os
    return os.listdir(path)


def new_script(deck_name):
    """
    script dictionary structure
    :param deck:
    :return:
    """
    # .strftime("%Y-%m-%d %H:%M:%S")
    current_time = datetime.datetime.now()
    script_dict = {'name': '',
                   'deck': deck_name,
                   'status': 'editing',
                   'prep': [],
                   'script': [],
                   'cleanup': [],
                   # 'time_created': current_time,
                   # 'time_modified': current_time,
                   # 'author': '',
                   }
    order = {'prep': [],
             'script': [],
             'cleanup': [],
             }
    return script_dict, order


def parse_functions(class_object=None, debug=False):
    functions = {}
    under_score = "_"
    if debug:
        under_score = "__"
    for function in dir(class_object):
        if not function.startswith(under_score) and not function.isupper():
            try:
                att = getattr(class_object, function)

                # handle getter setters
                if callable(att):
                    functions[function] = inspect.signature(att)
                else:
                    att = getattr(class_object.__class__, function)
                    if isinstance(att, property) and att.fset is not None:
                        setter = att.fset.__annotations__
                        setter.pop('return', None)
                        if setter:
                            functions[function] = setter
            except Exception:
                pass
    return functions


def is_compatible(att):
    try:
        obj = inspect.signature(att)
        try:
            pickle.dumps(obj)
        except Exception:
            return False
    except ValueError:
        return False
    return True


def config(script_dict):
    """
    take the global script_dict
    :return: list of variable that require input
    """
    configure = []
    for action in script_dict['script']:
        args = action['args']
        if args is not None:
            if type(args) is not dict:
                if type(args) is str and args.startswith("#") and not args[1:] in configure:
                    configure.append(args[1:])
            else:
                for arg in args:
                    if type(args[arg]) is str \
                            and args[arg].startswith("#") \
                            and not args[arg][1:] in configure:
                        configure.append(args[arg][1:])
    return configure


def config_return(script_dict):
    """
    take the global script_dict
    :return: list of variable that require input
    """
    return_list = [action['return'] for action in script_dict if not action['return'] == '']
    output_str = "return {"
    for i in return_list:
        output_str += "'" + i + "':" + i + ","
    output_str += "}"
    return output_str, return_list


def _get_type_from_parameters(arg, parameters):
    arg_type = ''
    if type(parameters) is inspect.Signature:
        p = parameters.parameters
        # print(p[arg].annotation)
        if p[arg].annotation is not inspect._empty:
            # print(p[arg].annotation)
            if p[arg].annotation.__module__ == 'typing':
                # print(p[arg].annotation.__args__)
                arg_type = [i.__name__ for i in p[arg].annotation.__args__]
            else:
                arg_type = p[arg].annotation.__name__
            # print(arg_type)
    elif type(parameters) is dict:
        if parameters[arg]:

            if parameters[arg].__module__ == 'typing':
                arg_type = [i.__name__ for i in parameters[arg].__args__]
            else:
                arg_type = parameters[arg].__name__
    return arg_type


def find_variable_in_script(script: Script, args: Dict[str, str]) -> Optional[Tuple[Dict[str, str], Dict[str, str]]]:
    # TODO: need to search for if the variable exists
    added_variables: list[Dict[str, str]] = [action for action in script.currently_editing_script if
                                             action["instrument"] == "variable"]

    possible_variable_arguments = {}
    possible_variable_types = {}

    for arg_name, arg_val in args.items():
        for added_variable in added_variables:
            if added_variable["action"] == arg_val:
                possible_variable_arguments[arg_name] = added_variable["action"]
                possible_variable_types[arg_name] = "variable"

    return possible_variable_arguments, possible_variable_types


def convert_type(args, parameters):
    bool_dict = {"True": True, "False": False}
    arg_types = {}
    if args:
        for arg in args:
            if args[arg] == '' or args[arg] == "None":
                args[arg] = None
                arg_types[arg] = _get_type_from_parameters(arg, parameters)
            elif args[arg] == "True" or args[arg] == "False":
                args[arg] = bool_dict[args[arg]]
                arg_types[arg] = 'bool'
            elif args[arg].startswith("#"):
                args[arg] = args[arg]
                arg_types[arg] = _get_type_from_parameters(arg, parameters)
            elif type(parameters) is inspect.Signature:
                p = parameters.parameters
                if p[arg].annotation is not inspect._empty:
                    if p[arg].annotation.__module__ == 'typing':
                        arg_types[arg] = p[arg].annotation.__args__

                        for i in p[arg].annotation.__args__:
                            try:
                                args[arg] = eval(f'{i}({args[arg]})')
                                break
                            except Exception:
                                pass

                    else:
                        args[arg] = p[arg].annotation(args[arg])
                        arg_types[arg] = p[arg].annotation.__name__
                else:
                    try:
                        args[arg] = eval(args[arg])
                        arg_types[arg] = ''
                    except Exception:
                        pass
            elif type(parameters) is dict:
                if parameters[arg]:
                    if parameters[arg].__module__ == 'typing':
                        # arg_types[arg] = parameters[arg].__args__
                        for i in parameters[arg].__args__:
                            # print(i)
                            try:
                                # args[arg] = i(args[arg])
                                args[arg] = eval(f'{i}({args[arg]})')
                                arg_types[arg] = i.__name__
                                break
                            except Exception:
                                pass
                    else:
                        args[arg] = parameters[arg](args[arg])
                        arg_types[arg] = parameters[arg].__name__
    return args, arg_types


def _convert_by_str(args, arg_types):
    # print(arg_types)
    if type(arg_types) is not list:
        arg_types = [arg_types]
    for i in arg_types:
        if i == "any":
            try:
                args = eval(args)
            except Exception:
                pass
            return args
        try:
            args = eval(f'{i}({args})')
            return args
        except Exception:
            pass
    raise TypeError(f"Input type error: cannot convert '{args}' to {i}.")


def _convert_by_class(args, arg_types):
    if arg_types.__module__ == 'builtins':
        args = arg_types(args)
        return args
    elif arg_types.__module__ == "typing":
        for i in arg_types.__args__:  # for typing.Union
            try:
                args = i(args)
                return args
            except Exception:
                pass
        raise TypeError("Input type error.")
    # else:
    #     args = globals()[args]
    return args


def convert_config_type(args, arg_types, is_class: bool = False):
    bool_dict = {"True": True, "False": False}
    # print(args, arg_types)
    # print(globals())
    if args:
        for arg in args:
            if arg not in arg_types.keys():
                raise ValueError("config file format not supported.")
            if args[arg] == '' or args[arg] == "None":
                args[arg] = None
            elif args[arg] == "True" or args[arg] == "False":
                args[arg] = bool_dict[args[arg]]
            else:
                arg_type = arg_types[arg]

                if is_class:
                    # if arg_type.__module__ == 'builtins':
                    args[arg] = _convert_by_class(args[arg], arg_type)
                else:
                    args[arg] = _convert_by_str(args[arg], arg_type)
    return args


def sort_actions(script_dict, order, script_type=None):
    """
    sort all three types if script_type is None, otherwise sort the specified script type
    :return:
    """
    if script_type:
        sort(script_dict, order, script_type)
    else:
        for i in stypes:
            sort(script_dict, order, i)


def sort(script_dict, order, script_type):
    if len(order[script_type]) > 0:
        for action in script_dict[script_type]:
            for i in range(len(order[script_type])):
                if action['id'] == int(order[script_type][i]):
                    # print(i+1)
                    action['id'] = i + 1
                    break
        order[script_type].sort()
        if not int(order[script_type][-1]) == len(script_dict[script_type]):
            new_order = list(range(1, len(script_dict[script_type]) + 1))
            order[script_type] = [str(i) for i in new_order]
        script_dict[script_type].sort(key=lambda x: x['id'])


def logic_dict(key: str, current_len, args, var_name=None):
    """

    :param key:
    :param current_len:
    :param args:
    :param var_name:
    :return:
    """
    logic_dict = {
        "if":
            [
                {"id": current_len + 1, "instrument": 'if', "action": 'if', "args": args, "return": ''},
                {"id": current_len + 2, "instrument": 'if', "action": 'else', "args": '', "return": ''},
                {"id": current_len + 3, "instrument": 'if', "action": 'endif', "args": '', "return": ''},
            ],
        "while":
            [
                {"id": current_len + 1, "instrument": 'while', "action": 'while', "args": args, "return": ''},
                {"id": current_len + 2, "instrument": 'while', "action": 'endwhile', "args": '', "return": ''},
            ],
        "variable":
            [
                {"id": current_len + 1, "instrument": 'variable', "action": var_name, "args": args, "return": ''},
            ]
    }
    return logic_dict[key]


# def make_grid(row:int=1,col:int=1):
#     """
#     return the tray index str list by defining the size
#     :param row: 1 to 26
#     :param col:
#     :return: return the tray index
#     """
#     letter_list = [chr(i) for i in range(65, 90)]
#     return [i + str(j + 1) for i in letter_list[:col] for j in range(row)]


def make_grid(row: int = 1, col: int = 1):
    """
    return the tray index str list by defining the size
    :param row: 1 to 26
    :param col: 1 to 26
    :return: return the tray index
    """
    letter_list = [chr(i) for i in range(65, 90)]
    return [[i + str(j + 1) for j in range(col)] for i in letter_list[:row]]


tray_size_dict = {
    "metal_4_6": {"row": 4, "col": 6},
    "metal_4_6_landscape": {"row": 6, "col": 4},
    "noah_hplc_tray": {"row": 4, "col": 7},
    "solvent_tray": {"row": 5, "col": 2},
}


def import_module_by_filepath(filepath: str, name: str):
    spec = importlib.util.spec_from_file_location(name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def if_deck_valid(module):
    count = 0
    for var in set(dir(module)):
        if not var.startswith("_") and not var[0].isupper() and not var.startswith("repackage") \
                and not type(eval("module." + var)).__module__ == 'builtins':
            count += 1
    return False if count == 0 else True


class SocketIOHandler(logging.Handler):
    def __init__(self, socketio: SocketIO):
        super().__init__()
        self.formatter = logging.Formatter('%(asctime)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        self.socketio = socketio

    def emit(self, record):
        message = self.format(record)
        # session["last_log"] = message
        self.socketio.emit('log', {'message': message})


def start_logger(socketio: SocketIO, logger_name: str, log_filename: str = None):
    # logging.basicConfig( format='%(asctime)s - %(message)s')
    formatter = logging.Formatter(fmt='%(asctime)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    file_handler = logging.FileHandler(filename='default.log', )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # console_logger = logging.StreamHandler()  # stream to console
    # logger.addHandler(console_logger)
    socketio_handler = SocketIOHandler(socketio)
    logger.addHandler(socketio_handler)
    return logger


def ax_wrapper(data):
    from ax.service.utils.instantiation import ObjectiveProperties
    parameter = []
    objectives = {}
    # Iterate through the webui_data dictionary
    for key, value in data.items():
        # Check if the key corresponds to a parameter type
        if "_type" in key:
            param_name = key.split("_type")[0]
            param_type = value
            param_value = data[f"{param_name}_value"].split(",")
            try:
                values = [float(v) for v in param_value]
            except Exception:
                values = param_value
            if param_type == "range":
                parameter.append({"name": param_name, "type": param_type, "bounds": values})
            if param_type == "choice":
                parameter.append({"name": param_name, "type": param_type, "values": values})
            if param_type == "fixed":
                parameter.append({"name": param_name, "type": param_type, "value": values[0]})
        elif key.endswith("_min"):
            if not value == 'none':
                obj_name = key.split("_min")[0]
                is_min = True if value == "minimize" else False

                threshold = None if f"{obj_name}_threshold" not in data else data[f"{obj_name}_threshold"]
                properties = ObjectiveProperties(minimize=is_min, threshold=threshold)
                objectives[obj_name] = properties
    return parameter, objectives


def ax_initiation(data):
    install_and_import("ax", "ax-platform")
    parameter, objectives = ax_wrapper(data)
    from ax.service.ax_client import AxClient
    ax_client = AxClient()
    ax_client.create_experiment(parameter, objectives)
    return ax_client


def get_arg_type(args, parameters):
    arg_types = {}
    # print(args, parameters)
    if args:
        for arg in args:
            arg_types[arg] = _get_type_from_parameters(arg, parameters)
    return arg_types


def install_and_import(package, package_name=None):
    try:
        # Check if the package is already installed
        importlib.import_module(package)
        # print(f"{package} is already installed.")
    except ImportError:
        # If not installed, install it
        # print(f"{package} is not installed. Installing now...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name or package])
        # print(f"{package} has been installed successfully.")


def process_data(data, config_type):
    rows = {}  # Dictionary to hold webui_data organized by rows

    # Organize webui_data by rows
    for key, value in data.items():
        if value:  # Only process non-empty values
            # Extract the field name and row index
            field_name, row_index = key.split('[')
            row_index = int(row_index.rstrip(']'))

            # If row not in rows, create a new dictionary for that row
            if row_index not in rows:
                rows[row_index] = {}

            # Add or update the field value in the specific row's dictionary
            rows[row_index][field_name] = value

    # Filter out any empty rows and create a list of dictionaries
    filtered_rows = [row for row in rows.values() if len(row) == len(config_type)]

    return filtered_rows



from flask import Blueprint, render_template, current_app
from flask_login import login_required

main = Blueprint('main', __name__)

@main.route("/")
@login_required
def index():
    off_line = current_app.config["OFF_LINE"]
    return render_template('home.html', off_line=off_line)


@main.route("/help")
def help_info():
    sample_deck = """
    from vapourtec.sf10 import SF10

    # connect SF10 pump
    sf10 = SF10(device_port="com7")

    # start ivoryOS
    from ivoryos.app import ivoryos
    ivoryos(__name__)
    """
    return render_template('help.html', sample_deck=sample_deck)



from flask import Blueprint, redirect, url_for, flash, request, render_template, session
from flask_login import login_required

from ivoryos.utils.db_models import Script, User, db
from ivoryos.utils.utils import get_script_file, post_script_file
global deck

database = Blueprint('database', __name__)


@database.route("/delete/<id>")
@login_required
def delete_action(id):
    back = request.referrer
    script = get_script_file()
    script.delete_action(id)
    post_script_file(script)
    return redirect(back)


@database.route("/edit_workflow/<workflow_name>")
@login_required
def edit_workflow(workflow_name):
    row = Script.query.get(workflow_name)
    script = Script(**row.as_dict())
    post_script_file(script)
    pseudo_name = session.get("pseudo_deck", "")
    if not deck and pseudo_name and not script.deck == pseudo_name:
        flash(f"Choose the deck with name {script.deck}")
    return redirect(url_for('design.experiment_builder'))


@database.route("/delete_workflow/<workflow_name>")
@login_required
def delete_workflow(workflow_name):
    Script.query.filter(Script.name == workflow_name).delete()
    db.session.commit()
    return redirect(url_for('database.load_from_database'))


@database.route("/publish")
@login_required
def publish():
    script = get_script_file()
    if not script.name or not script.deck:
        flash("Deck cannot be empty, try to re-submit deck configuration on the left panel")
    row = Script.query.get(script.name)
    if row and row.status == "finalized":
        flash("This is a protected script, use save as to rename.")
    elif row and not session['user'] == row.author:
        flash("You are not the author, use save as to rename.")
    else:
        db.session.merge(script)
        db.session.commit()
        flash("Saved!")
    return redirect(url_for('design.experiment_builder'))


@database.route("/finalize")
@login_required
def finalize():
    script = get_script_file()
    script.finalize()
    db.session.merge(script)
    db.session.commit()
    post_script_file(script)
    return redirect(url_for('design.experiment_builder'))


@database.route("/database/", methods=['GET', 'POST'])
@database.route("/database/<deck_name>", methods=['GET', 'POST'])
@login_required
def load_from_database(deck_name=None):
    session.pop('edit_action', None)  # reset cache
    query = Script.query
    search_term = request.args.get("keyword", None)
    if search_term:
        query = query.filter(Script.name.like(f'%{search_term}%'))
    if deck_name is None:
        temp = Script.query.with_entities(Script.deck).distinct().all()
        deck_list = [i[0] for i in temp]
    else:
        query = query.filter(Script.deck == deck_name)
        deck_list = ["ALL"]
    page = request.args.get('page', default=1, type=int)
    per_page = 10

    workflows = query.paginate(page=page, per_page=per_page, error_out=False)
    return render_template("experiment_database.html", workflows=workflows, deck_list=deck_list, deck_name=deck_name)


@database.route("/edit_run_name", methods=['GET', 'POST'])
@login_required
def edit_run_name():
    if request.method == "POST":
        run_name = request.form.get("run_name")
        exist_script = Script.query.get(run_name)
        if not exist_script:
            script = get_script_file()
            script.save_as(run_name)
            post_script_file(script)
        else:
            flash("Script name is already exist in database")
        return redirect(url_for("design.experiment_builder"))


@database.route("/save_as", methods=['GET', 'POST'])
@login_required
def save_as():
    # script = get_script_file()
    if request.method == "POST":
        run_name = request.form.get("run_name")
        exist_script = Script.query.get(run_name)
        if not exist_script:
            script = get_script_file()
            script.save_as(run_name)
            script.author = session.get('user')
            post_script_file(script)
            publish()
        else:
            flash("Script name is already exist in database")
        return redirect(url_for("design.experiment_builder"))



import os
import pickle
import sys

from flask import Blueprint, redirect, url_for, flash, request, render_template, session, current_app
from flask_login import login_required

from ivoryos.utils.global_config import GlobalConfig
from ivoryos.utils import utils
from ivoryos.utils.form import create_form_from_module
from ivoryos.utils.utils import get_script_file

global_config = GlobalConfig()

control = Blueprint('control', __name__)

global deck
deck = None


@control.route("/my_deck")
@login_required
def deck_controllers():
    global deck
    if deck is None:
        # print("loading deck")
        module = current_app.config.get('MODULE', '')
        deck = sys.modules[module] if module else None
        # global_config.deck = deck
    deck_variables = parse_deck(deck)
    # deck_variables = session.get("deck_variables")
    return render_template('controllers_home.html', defined_variables=deck_variables, deck=True)


@control.route("/new_controller/")
@control.route("/new_controller/<instrument>", methods=['GET', 'POST'])
@login_required
def new_controller(instrument=None):
    device = None
    args = None
    if instrument:

        device = find_instrument_by_name(instrument)
        args = utils.inspect.signature(device.__init__)

        if request.method == 'POST':
            device_name = request.form.get("device_name", "")
            if device_name and device_name in globals():
                flash("Device name is defined. Try another name, or leave it as blank to auto-configure")
                return render_template('controllers_new.html', instrument=instrument,
                                       api_variables=global_config.api_variables,
                                       device=device, args=args, defined_variables=global_config.defined_variables)
            if device_name == "":
                device_name = device.__name__.lower() + "_"
                num = 1
                while device_name + str(num) in global_config.defined_variables:
                    num += 1
                device_name = device_name + str(num)
            kwargs = request.form.to_dict()
            kwargs.pop("device_name")
            for i in kwargs:
                if kwargs[i] in globals():
                    kwargs[i] = globals()[kwargs[i]]
            try:
                utils.convert_config_type(kwargs, device.__init__.__annotations__, is_class=True)
            except Exception as e:
                flash(e)
            try:
                global_config.defined_variables[device_name] = device(**kwargs)
                # global_config.defined_variables.add(device_name)
                return redirect(url_for('control.controllers_home'))
            except Exception as e:
                flash(e)
    return render_template('controllers_new.html', instrument=instrument, api_variables=global_config.api_variables,
                           device=device, args=args, defined_variables=global_config.defined_variables)


@control.route("/controllers")
@login_required
def controllers_home():
    # defined_variables = parse_deck(deck)
    return render_template('controllers_home.html', defined_variables=global_config.defined_variables)


@control.route("/controllers/<instrument>", methods=['GET', 'POST'])
@login_required
def controllers(instrument):
    global deck
    if instrument.startswith("deck") and deck is None:
        module = current_app.config.get('MODULE', '')
        deck = sys.modules[module] if module else None
    inst_object = find_instrument_by_name(instrument)
    _forms = create_form_from_module(sdl_module=find_instrument_by_name(instrument), autofill=False, design=False)
    card_order = session.get('card_order')
    order = card_order.get(instrument, _forms.keys())
    if instrument not in card_order:
        card_order[instrument] = list(order)
        session['card_order'] = card_order
        # print(session['card_order'])
    forms = {name: _forms[name] for name in order if name in _forms}
    if request.method == 'POST':
        all_kwargs = request.form.copy()
        method_name = all_kwargs.pop("hidden_name", None)
        # if method_name is not None:
        form = forms.get(method_name)
        kwargs = {field.name: field.data for field in form if field.name != 'csrf_token'}
        function_executable = getattr(inst_object, method_name)
        if form and form.validate_on_submit():
            try:
                kwargs.pop("hidden_name")
                output = function_executable(**kwargs)
                flash(f"\nRun Success! Output value: {output}.")
            except Exception as e:
                flash(e.__str__())
        else:
            flash(form.errors)
    return render_template('controllers.html', instrument=instrument, inst=inst_object, forms=forms)


@control.route("/import_api", methods=['GET', 'POST'])
def import_api():
    filepath = request.form.get('filepath')
    # filepath.replace('\\', '/')
    name = os.path.split(filepath)[-1].split('.')[0]
    try:
        spec = utils.importlib.util.spec_from_file_location(name, filepath)
        module = utils.importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        classes = utils.inspect.getmembers(module, utils.inspect.isclass)
        if len(classes) == 0:
            flash("Invalid import: no class found in the path")
            return redirect(url_for("control.controllers_home"))
        for i in classes:
            globals()[i[0]] = i[1]
            global_config.api_variables.add(i[0])
    # should handle path error and file type error
    except Exception as e:
        flash(e.__str__())
    return redirect(url_for("control.new_controller"))


@control.route("/disconnect", methods=["GET"])
@control.route("/disconnect/<device_name>", methods=["GET"])
def disconnect(device_name=None):
    """TODO handle disconnect device"""
    if device_name:
        try:
            exec(device_name + ".disconnect()")
        except Exception:
            pass
        global_config.defined_variables.remove(device_name)
        globals().pop(device_name)
        return redirect(url_for('control.controllers_home'))

    deck_variables = ["deck." + var for var in set(dir(deck))
                      if not (var.startswith("_") or var[0].isupper() or var.startswith("repackage"))
                      and not type(eval("deck." + var)).__module__ == 'builtins']
    for i in deck_variables:
        try:
            exec(i + ".disconnect()")
        except Exception:
            pass
    globals()["deck"] = None
    return redirect(url_for('control.deck_controllers'))


@control.route("/import_deck", methods=['POST'])
def import_deck():
    global deck
    script = get_script_file()
    filepath = request.form.get('filepath')
    session['dismiss'] = request.form.get('dismiss')
    update = request.form.get('update')
    back = request.referrer
    if session['dismiss']:
        return redirect(back)
    # if filepath == "manage history":

    name = os.path.split(filepath)[-1].split('.')[0]
    try:
        module = utils.import_module_by_filepath(filepath=filepath, name=name)
        # deck format checking
        if not utils.if_deck_valid(module):
            flash("Invalid Deck import")
            return redirect(url_for("control.deck_controllers"))
        globals()["deck"] = module
        utils.save_to_history(filepath, current_app.config["DECK_HISTORY"])
        parse_deck(deck, save=update)

        if script.deck is None:
            script.deck = module.__name__
    # file path error exception
    except Exception as e:
        flash(e.__str__())
    return redirect(back)


@control.route('/save-order/<instrument>', methods=['POST'])
def save_order(instrument):
    # Save the new order for the specified group to session
    data = request.json
    card_order = session.get("card_order", {})
    card_order[instrument] = data['order']
    session['card_order'] = card_order
    return '', 204


@control.route('/hide_function/<instrument>/<function>')
def hide_function(instrument, function):
    back = request.referrer
    hidden_functions = session.get("hidden_functions")
    functions = hidden_functions.get(instrument, [])
    card_order = session.get("card_order")
    order = card_order.get(instrument)
    if function not in functions:
        functions.append(function)
        order.remove(function)
    hidden_functions[instrument] = functions
    card_order[instrument] = order
    session['hidden_functions'] = hidden_functions
    session['card_order'] = card_order
    return redirect(back)


@control.route('/remove_hidden/<instrument>/<function>')
def remove_hidden(instrument, function):
    back = request.referrer
    hidden_functions = session.get("hidden_functions")
    functions = hidden_functions.get(instrument, [])
    card_order = session.get("card_order")
    order = card_order.get(instrument)
    if function in functions:
        functions.remove(function)
        order.append(function)
    hidden_functions[instrument] = functions
    card_order[instrument] = order
    session['hidden_functions'] = hidden_functions
    session['card_order'] = card_order
    return redirect(back)


def parse_deck(deck, save=None):
    parse_dict = {}
    # TODO
    deck_variables = ["deck." + var for var in set(dir(deck))
                      if not (var.startswith("_") or var[0].isupper() or var.startswith("repackage"))
                      and not type(eval("deck." + var)).__module__ == 'builtins'
                      ]
    session["deck_variables"] = deck_variables
    for var in deck_variables:
        instrument = eval(var)
        functions = utils.parse_functions(instrument)
        parse_dict[var] = functions
    if deck is not None and save:
        # pseudo_deck = parse_dict
        parse_dict["deck_name"] = os.path.splitext(os.path.basename(deck.__file__))[
            0] if deck.__name__ == "__main__" else deck.__name__
        with open(os.path.join(current_app.config["DUMMY_DECK"], f"{parse_dict['deck_name']}.pkl"), 'wb') as file:
            pickle.dump(parse_dict, file)
    return deck_variables


def find_instrument_by_name(name: str):
    if name.startswith("deck"):
        return eval(name)
    elif name in global_config.defined_variables:
        return global_config.defined_variables[name]
    elif name in globals():
        return globals()[name]



import inspect
import json
import os
import re

# import ollama
from openai import OpenAI


# load_dotenv()

# host = "137.82.65.246"
# model = "llama3"


class LlmAgent:
    def __init__(self, model="llama3", output_path=os.curdir, host=None):
        self.host = host
        self.base_url = f"http://{self.host}:11434/v1/" if host is not None else ""
        self.model = model
        self.output_path = os.path.join(output_path, "llm_output")
        self.client = OpenAI(api_key=os.getenv("OPENAI_API_KEY")) if host is None else OpenAI(api_key="ollama",
                                                                                              base_url=self.base_url)
        os.makedirs(self.output_path, exist_ok=True)

    def extract_annotations_docstrings(self, cls):
        extracted_methods = []
        for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            if not name.startswith('_'):
                annotation = inspect.signature(method).return_annotation
                docstring = inspect.getdoc(method)
                extracted_methods.append((name, annotation, docstring))
        method_name = [t[0] for t in extracted_methods]
        variables = [(key, value) for key, value in vars(cls).items() if
                     not key.startswith("_") and not key in method_name]
        class_str = ""
        # class_str = f"class {cls.__name__}:\n"
        # class_str += f'\t"""{inspect.getdoc(cls)}"""\n\n'
        for key, value in variables:
            class_str += f'\t{key}={value}\n'
        for name, annotation, docstring in extracted_methods:
            class_str += f'\tdef {name}{inspect.signature(getattr(cls, name))}:\n'
            class_str += f'\t\t"""\n{docstring}\n"""' + '\n\n' if docstring else ''
        class_str = class_str.replace('self, ', '')
        class_str = class_str.replace('self', '')
        name_list = [name for name, _, _ in extracted_methods]
        # print(class_str)
        current_path = os.path.curdir
        with open(os.path.join(self.output_path, "docstring_manual.txt"), "w") as f:
            f.write(class_str)

        return class_str, name_list

    @staticmethod
    def parse_code_from_msg(msg):
        msg = msg.strip()
        # print(msg)
        # code_blocks = re.findall(r'```(?:json\s)?(.*?)```', msg, re.DOTALL)
        code_blocks = re.findall(r'\[\s*\{.*?\}\s*\]', msg, re.DOTALL)

        json_blocks = []
        for block in code_blocks:
            if not block.startswith('['):
                start_index = block.find('[')
                block = block[start_index:]
            block = re.sub(r'//.*', '', block)
            block = block.replace('True', 'true').replace('False', 'false')
            try:
                # Try to parse the block as JSON
                json_data = json.loads(block.strip())
                if isinstance(json_data, list):
                    json_blocks = json_data
            except json.JSONDecodeError:
                continue
        return json_blocks

    def _generate(self, robot, prompt):
        deck_info, name_list = self.extract_annotations_docstrings(type(robot))
        full_prompt = '''
                                I have some python functions, for example when calling them I want to write them using JSON, 
                                it is necessary to include all args
                                for example
                                def dose_solid(amount_in_mg:float, bring_in:bool=True): def analyze():
                                dose_solid(3)
                                analyze()
                                I would want to write to
                                [
                                {
                                    "action": "dose_solid",
                                    "arg_types": {
                                        "amount_in_mg": "float",
                                        "bring_in": "bool"
                                    },
                                    "args": {
                                        "amount_in_mg": 3,
                                        "bring_in": true
                                    }
                                },
                                {
                                    "action": "analyze",
                                    "arg_types": {},
                                    "args": {}
                                }
                                ]

                                ''' + f'''
                                Now these are my callable functions,
                                {deck_info}

                                and I want you to find the most appropriate function if I want to do these tasks
                                """{prompt}"""
                                ,and write a list of dictionary in json accordingly. Please only use these action names {name_list}, 
                                can you also help find the default value you can't find the info from my request.
                                '''

        with open(os.path.join(self.output_path, "prompt.txt"), "w") as f:
            f.write(full_prompt)
        messages = [{"role": "user",
                     "content": full_prompt}]
        # if self.host == "openai":
        output = self.client.chat.completions.create(
            messages=messages,
            model=self.model,
        )
        msg = output.choices[0].message.content

        # else:
        #     output = self.client.chat(
        #         model=self.model,
        #         messages=messages,
        #         # stream=False,
        #     )
        #     msg = output['message']['content']
        # print(msg)

        code = self.parse_code_from_msg(msg)
        code = [action for action in code if action['action'] in name_list]
        # print('\033[91m', code, '\033[0m')
        return code

    def generate_code(self, robot, prompt, attempt_allowance: int = 3):
        attempt = 0
        while attempt < attempt_allowance:
            code = self._generate(robot, prompt)
            attempt += 1
            if code:
                break

        # print(attempt)
        return code


if __name__ == "__main__":
    from example.sdl_example.abstract_sdl import deck

    # llm_agent = LlmAgent(host="openai", model="gpt-3.5-turbo")
    llm_agent = LlmAgent(host="localhost", model="llama3.1")
    # robot = IrohDeck()
    # extract_annotations_docstrings(DummySDLDeck)
    prompt = '''I want to start with dosing 10 mg of current sample, and add 1 mL of toluene 
    and equilibrate for 10 minute at 40 degrees, then sample 20 ul of sample to analyze with hplc, and save result'''
    code = llm_agent.generate_code(deck, prompt)
    print(code)
"""
I want to dose 10mg, 6mg, 4mg, 3mg, 2mg, 1mg to 6 vials
I want to add 10 mg to vial a3, and 10 ml of liquid, then shake them for 3 minutes

"""



from setuptools import setup, find_packages

setup(
    name='ivoryOS',
    version='0.1.4',
    packages=find_packages(exclude=['example', 'example.*']),
    include_package_data=True,
    description='an open-source Python package enabling Self-Driving Labs (SDLs) interoperability',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Ivory Zhang',
    author_email='ivoryzhang@chem.ubc.ca',
    license='MIT',
    install_requires=[
        # "ax-platform",
        "bcrypt",
        "Flask-Login",
        "Flask-Session",
        "Flask-SocketIO",
        "Flask-SQLAlchemy",
        "Flask-WTF",
        "SQLAlchemy-Utils",
        # "openai",
        "python-dotenv",
    ],
    url='https://gitlab.com/heingroup/ivoryos'
)



import os
import csv
import threading
import time
from datetime import datetime

from ivoryos.utils import utils
from ivoryos.utils.db_models import Script


class ScriptRunner:
    def __init__(self, globals_dict=None):
        if globals_dict is None:
            globals_dict = {}
        self.globals_dict = globals_dict

        self.stop_event = threading.Event()
        self.is_running = False
        self.lock = threading.Lock()

    def reset_stop_event(self):
        self.stop_event.clear()

    def stop_execution(self):
        self.stop_event.set()
        # print("Stop pending tasks")

    def run_script(self, script, repeat_count=1, run_name=None, logger=None, socketio=None, config=None, bo_args=None,
                   output_path=""):
        time.sleep(1)
        with self.lock:
            if self.is_running:
                logger.info("System is busy. Please wait for it to finish or stop it before starting a new one.")
                return None
            self.is_running = True

        self.reset_stop_event()

        thread = threading.Thread(target=self._run_with_stop_check,
                                  args=(script, repeat_count, run_name, logger, socketio, config, bo_args, output_path))
        thread.start()
        return thread

    def _run_with_stop_check(self, script: Script, repeat_count, run_name, logger, socketio, config, bo_args,
                             output_path):
        time.sleep(1)
        try:
            # Run "prep" section once
            script_dict = script.script_dict
            self._run_actions(script_dict.get("prep", []), section_name="prep", run_name=run_name, logger=logger)
            output_list = []
            _, arg_type = script.config("script")
            _, return_list = script.config_return()

            # Run "script" section multiple times
            if repeat_count:
                self._run_repeat_section(repeat_count, bo_args, output_list, return_list, run_name, logger, socketio)
            elif config:
                self._run_config_section(config, arg_type, output_list, return_list, run_name, logger, socketio)

            # Run "cleanup" section once

            self._run_actions(script_dict.get("cleanup", []), section_name="cleanup", run_name=run_name, logger=logger)

            # Save results if necessary
            if output_list:
                self._save_results(run_name, arg_type, return_list, output_list, logger, output_path)
        except Exception as e:
            logger.error(f"Error during script execution: {e}")
        finally:
            with self.lock:
                self.is_running = False  # Reset the running flag when done

    def _run_actions(self, actions, section_name="", run_name=None, logger=None):
        for action in actions:
            if self.stop_event.is_set():
                logger.info(f"Stopping execution during {section_name} section.")
                break
            logger.info(f'Executing {section_name} steps')
            fname = f"{run_name}_{section_name}"
            function = self.globals_dict[fname]
            function()

    def _run_config_section(self, config, arg_type, output_list, return_list, run_name, logger, socketio):
        compiled = True
        for i in config:
            try:
                i = utils.convert_config_type(i, arg_type)
            except Exception as e:
                logger.info(e)
                compiled = False
                break
        if compiled:
            for i, kwargs in enumerate(config):
                kwargs = dict(kwargs)
                logger.info(f'Executing {i + 1} of {len(config)} with kwargs = {kwargs}')
                progress = (i + 1) * 100 / len(config)
                socketio.emit('progress', {'progress': progress})
                fname = f"{run_name}_script"
                function = self.globals_dict[fname]
                output = function(**kwargs)
                if output:
                    kwargs.update(output)
                    output_list.append(kwargs)

    def _run_repeat_section(self, repeat_count, bo_args, output_list, return_list, run_name, logger, socketio):
        if bo_args:
            logger.info('Initializing optimizer...')
            ax_client = utils.ax_initiation(bo_args)
        for i in range(int(repeat_count)):
            if self.stop_event.is_set():
                logger.info(f'Stopping execution during {run_name}: {i + 1}/{int(repeat_count)}')
                break
            logger.info(f'Executing {run_name}: {i + 1}/{int(repeat_count)}')
            progress = (i + 1) * 100 / int(repeat_count)
            socketio.emit('progress', {'progress': progress})
            if bo_args:
                try:
                    parameters, trial_index = ax_client.get_next_trial()
                    logger.info(f'Output value: {parameters}')
                    fname = f"{run_name}_script"
                    function = self.globals_dict[fname]
                    output = function(**parameters)
                    # output = eval(f"{run_name}_script(**{parameters})")
                    ax_client.complete_trial(trial_index=trial_index, raw_data=output)
                except Exception as e:
                    logger.info(f'Optimization error: {e}')
                    break
            else:
                fname = f"{run_name}_script"
                function = self.globals_dict[fname]
                output = function()

            if output:
                output_list.append(output)
                logger.info(f'Output value: {output}')

    @staticmethod
    def _save_results(run_name, arg_type, return_list, output_list, logger, output_path):
        if arg_type:
            args = list(arg_type.keys())
            args.extend(return_list)
            filename = run_name + "_" + datetime.now().strftime("%Y-%m-%d %H-%M") + ".csv"
            file_path = os.path.join(output_path, filename)
            with open(file_path, "w", newline='') as file:
                writer = csv.DictWriter(file, fieldnames=args)
                writer.writeheader()
                writer.writerows(output_list)
            logger.info(f'Results saved to {file_path}')



import os
from typing import Union

from flask import Flask

from ivoryos.config import Config, get_config
from ivoryos.routes.auth import auth, login_manager
from ivoryos.routes.control import control
from ivoryos.routes.database import database
from ivoryos.routes.design import design, socketio
from ivoryos.routes.main import main
from ivoryos.utils import utils
from ivoryos.utils.db_models import db
from ivoryos.utils.global_config import GlobalConfig

global_config = GlobalConfig()


def create_app(config_class=None):
    app = Flask(__name__)
    app.config.from_object(config_class or 'config.get_config()')

    # Initialize extensions
    socketio.init_app(app)
    login_manager.init_app(app)
    login_manager.login_view = "auth.login"
    db.init_app(app)

    # Create database tables
    with app.app_context():
        db.create_all()

    # Additional setup
    utils.create_gui_dir(app.config['OUTPUT_FOLDER'])

    # logger_list = app.config["LOGGERS"]
    logger_path = os.path.join(app.config["OUTPUT_FOLDER"], app.config["LOGGERS_PATH"])
    logger = utils.start_logger(socketio, 'gui_logger', logger_path)

    @app.before_request
    def before_request():
        from flask import g
        g.logger = logger
        g.socketio = socketio

    app.register_blueprint(main)
    app.register_blueprint(auth)
    app.register_blueprint(design)
    app.register_blueprint(database)
    app.register_blueprint(control)

    return app


def ivoryos(module=None, host="0.0.0.0", port=None, debug=None, llm_server=None, model=None,
            config: Config = None,
            logger: Union[str, list] = None,
            logger_output_name: str = None,
            ):
    app = create_app(config_class=config or get_config())  # Create app instance using factory function

    port = port or int(os.environ.get("PORT", 8000))
    debug = debug if debug is not None else app.config.get('DEBUG', True)

    app.config["LOGGERS"] = logger
    app.config["LOGGERS_PATH"] = logger_output_name or app.config["LOGGERS_PATH"]
    logger_path = os.path.join(app.config["OUTPUT_FOLDER"], app.config["LOGGERS_PATH"])

    if module:
        app.config["MODULE"] = module
        app.config["OFF_LINE"] = False
    else:
        app.config["OFF_LINE"] = True
    if model:
        app.config["ENABLE_LLM"] = True
        app.config["LLM_MODEL"] = model
        app.config["LLM_SERVER"] = llm_server
        utils.install_and_import('openai')
        from ivoryos.utils.llm_agent import LlmAgent
        global_config.agent = LlmAgent(host=llm_server, model=model,
                                       output_path=os.path.dirname(os.path.abspath(module)))
    else:
        app.config["ENABLE_LLM"] = False
    if logger and type(logger) is str:
        utils.start_logger(socketio, log_filename=logger_path, logger_name=logger)
    elif type(logger) is list:
        for log in logger:
            utils.start_logger(socketio, log_filename=logger_path, logger_name=log)
    socketio.run(app, host=host, port=port, debug=debug, use_reloader=False, allow_unsafe_werkzeug=True)



from flask import Blueprint, redirect, url_for, flash, request, render_template, session
from flask_login import login_required, login_user, logout_user, LoginManager
import bcrypt

from ivoryos.utils.db_models import Script, User, db
from ivoryos.utils.utils import post_script_file
login_manager = LoginManager()

auth = Blueprint('auth', __name__)


@auth.route('/login', methods=['GET', 'POST'])
def login():
    if request.method == 'POST':
        username = request.form.get('username')
        password = request.form.get('password')

        # session.query(User, User.name).all()
        user = db.session.query(User).filter(User.username == username).first()
        input_password = password.encode('utf-8')
        # if user and bcrypt.checkpw(input_password, user.hashPassword.encode('utf-8')):
        if user and bcrypt.checkpw(input_password, user.hashPassword):
            # password.encode("utf-8")
            # user = User(username, password.encode("utf-8"))
            login_user(user)
            session['user'] = username
            script_file = Script(author=username)
            session["script"] = script_file.as_dict()
            session['hidden_functions'], session['card_order'], session['prompt'] = {}, {}, {}
            session['autofill'] = False
            post_script_file(script_file)
            return redirect(url_for('main.index'))
        else:
            flash("Incorrect username or password")
    return render_template('login.html')


@auth.route('/signup', methods=['GET', 'POST'])
def signup():
    if request.method == 'POST':
        username = request.form.get('username')
        password = request.form.get('password')
        salt = bcrypt.gensalt()
        hashed = bcrypt.hashpw(password.encode('utf-8'), salt)
        user = User(username, hashed)
        try:
            db.session.add(user)
            db.session.commit()
            return redirect(url_for('auth.login'))
        except Exception:
            flash("username exists :(")
    return render_template('signup.html')


@auth.route("/logout")
@login_required
def logout():
    logout_user()
    session.clear()
    return redirect(url_for('auth.login'))


@login_manager.user_loader
def load_user(username):
    return User(username, password=None)
